!+
! Module: emldetect_mod
!
! Authors:  Hermann Brunner - AIP
!          
! Description: Maximum likelihood EPIC source searching
! Comments:
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module emldetect_mod

  use param  
  use dal
  use edetect_io_ml
  use emldata
  use emlfit
  use eml_io
  use psf


  implicit none 
  
  private 

  public emldetect

contains 
  
  !---------------------------------------------------------------------
  subroutine cut_radius (x,y,scut2,ecut2,enfrac)
  !---------------------------------------------------------------------

    use cal

    real, intent(in)  :: x, y
    real, intent(out) :: scut2, ecut2
    real, dimension(max_ima), intent(out) :: enfrac 

    real, parameter :: sec2mm = 7500.0*pi180/3600.0
    integer :: i,j,k,ipoint,ierr
    real    :: scut0,ecut0,ecut1,scut,ecut
    real(kind=double) :: E, theta, fraction 
    logical :: first = .true. 

    type (PsfT) :: ppsf
    type(PsfValidityRangesT) :: vrange 

!!! for testing
!    type(PsfBinSizeT), dimension(max_ima)  :: bsiz
!    type(PSFwindowT)   :: window 
!    real, dimension(:,:), pointer :: pima
!    integer :: nx,ny,istate1
!!!!   
    save

    ierr=0
    
    scut=0.
    ecut=0.
    
      if ((ences .lt. 1.0) .or. (encee .lt. 1.0)) then

!      always call setState, in case values are overwritten elsewhere (30/7/04)     
!      if (first) then
        call CAL_setState1(instrument=EPN, &
                         AccuracyLevel=ACCURACY_MEDIUM)
!      call CAL_setState1(instrument=EMOS1,AccuracyLevel=ACCURACY_MEDIUM)

        vrange = CAL_PsfValidityRanges()
        first = .false.
!      end if



        do i=1,num_instr
          theta=sqrt((x-xcen(i))**2+(y-ycen(i))**2)*pixim 
          theta = max(min(theta,vrange%thetaMax-1.0),vrange%thetaMin+1.0)
          do j=1,num_image(i)
            ipoint = ip(i,j) 
            E=energy(ipoint)
            ppsf = Cal_getPSF(E, theta, 0d0) 
            fraction=ences
            if (fraction .lt. 1.0) then 
              scut0 = CAL_psfEncircledenergy(ppsf,fraction)/(pixim*sec2mm)
            else
              scut0 = ences
            endif
            scut=max(scut,scut0)
            fraction=encee
            if (fraction .lt. 1.0) then 
              ecut0 = CAL_psfEncircledenergy(ppsf,fraction)/(pixim*sec2mm)
            else
              ecut0 = encee
            endif
            ecut=max(ecut,ecut0)
          end do
          do j=1,num_image(i)
            ipoint = ip(i,j) 
            E=energy(ipoint)
            ppsf = Cal_getPSF(E, theta, 0d0) 
            ecut1=0. 
            do k=1,99
              fraction=k*0.01
              ecut0=ecut1 
              ecut1=CAL_psfEncircledenergy(ppsf,fraction)/(pixim*sec2mm)
              if (ecut1>ecut) exit
            end do
            enfrac(j)=((k-1)*(ecut1-ecut)+k*(ecut-ecut0))*0.01/(ecut1-ecut0)
          end do
        end do
      endif      

     ! if parameter value >= 1.0, interprete input in units of image pixels

     if (encee .ge. 1.0) ecut=encee
     if (ences .ge. 1.0) scut=ences

     scut2 = scut*scut
     ecut2 = ecut*ecut



  end subroutine cut_radius

  !---------------------------------------------------------------------
  subroutine sou_clus(xms,yms,scts,ext,insrc,nsrc,scut2,xx,yy,src,extent,in, &
                      k,nsou,indx,iwork,threshold, selvalue) 
  !---------------------------------------------------------------------

    real, dimension(max_sou), intent(in)  :: xx, yy, extent, selvalue 
    real, dimension(max_ima, max_sou), intent(in) :: src
    real, intent(in) :: scut2
    integer, dimension(max_sou), intent(in) :: in, indx
    integer, intent(in) :: k, nsou

    integer, dimension(max_sou), intent(inout) :: iwork

    real, dimension(max_fit), intent(out) :: xms, yms, ext
    real, dimension(max_ima,max_fit), intent(out) :: scts   
    integer, dimension(max_fit), intent(out) :: insrc
    integer, intent(out) :: nsrc
    real, dimension(max_fit) :: value
    real, intent(in) :: threshold
    integer :: nmulcount
    real, dimension(max_ima,max_fit) :: scts_old
    real, dimension(max_fit) :: xms_old,yms_old,ext_old
    integer, dimension(max_fit) :: insrc_old

    integer :: i,j,kk,sign,m
    logical :: found      

    insrc(:)=0

    kk=indx(nsou-k+1)
    xms(1) = xx(kk)
    yms(1) = yy(kk)
    ext(1) = extent(kk)
    scts(1:nim,1)=src(1:nim,kk)
    insrc(1)=in(kk)
    value(1)=selvalue(kk)

    nsrc=1
    found = .true.

    do while (found.and.nsrc.lt.max_sfit)
      found = .false.
      do j=k+1,nsou
        kk=indx(nsou-j+1)
        if(iwork(kk).eq.0) then
          do i=1,nsrc
            if ((xms(nsrc)-xx(kk))**2+(yms(nsrc)-yy(kk))**2 &
                .le.scut2.and.nsrc.lt.max_sfit) then
              found = .true.
              nsrc=nsrc+1
              iwork(kk)=1        
              xms(nsrc)=xx(kk)
              yms(nsrc)=yy(kk)
              value(nsrc)=selvalue(kk)
              ext(nsrc) = extent(kk)
              scts(1:nim,nsrc)=src(1:nim,kk)
              insrc(nsrc)=in(kk)
              exit
            end if
          end do  
        endif
        if (found) exit              
      end do   
    end do
!   write(6,*)' source ',in(k),' has ',nsrc-1,' satellites ',max_sfit

!   write(*,*) "before nmulsou:", insrc

!    if (nsmul>1) then
!      nmulcount=0
!      do i=nsrc,1,-1
!        if (value(i) > threshold) then
!        nmulcount=nmulcount+1
!        xms(nsmul*i)=xms(i)
!        yms(nsmul*i)=yms(i)
!        scts(1:nim,nsmul*i)=scts(1:nim,i)/nsmul
!        insrc(i*nsmul)=insrc(i)
!        do j=1,nsmul-1
!          xms(nsmul*i-j)=xms(i)
!          yms(nsmul*i-j)=yms(i)
!          ! first guess for additional source is point-like 
!          ext(nsmul*i-j)=0.0
!          scts(1:nim,nsmul*i-j)=scts(1:nim,i)/nsmul
!          insrc(i*nsmul-j)=insrc(i)
!        end do
!        
!        endif
!      enddo
!!      nsrc=nsmul*nsrc
!       nsrc=nsrc+(nsmul-1)*nmulcount
!    end if 


    if (nsmul>1) then
      xms_old=xms
      yms_old=yms
      scts_old(1:nim,:)=scts(1:nim,:)
      ext_old=ext
      insrc_old=insrc
      nmulcount=0
      do i=1,nsrc
        if (value(i) > threshold .or. .not. threshfla) then
          do j=1,nsmul
            nmulcount=nmulcount+1
!            xms(nmulcount)=xms_old(i)
!            yms(nmulcount)=yms_old(i)
! test modifying input positions:
            sign=1
            do m=1,j
              sign=sign*(-1)
            enddo
            xms(nmulcount)=xms_old(i)+min(float(j-1),2.0)*sign
            sign=-1
            do m=1,j
              sign=sign*(-1)
            enddo
            yms(nmulcount)=yms_old(i)+min(float(j-1),2.0)*sign

            ! first guess for additional source is point-like 
            ext(nmulcount)=0.0
            scts(1:nim,nmulcount)=scts_old(1:nim,i)/nsmul
            insrc(nmulcount)=insrc_old(i)
          end do
        else
          nmulcount=nmulcount+1
          xms(nmulcount)=xms_old(i)
          yms(nmulcount)=yms_old(i)
          ext(nmulcount)=ext_old(i)
          scts(1:nim,nmulcount)=scts_old(1:nim,i)
          insrc(nmulcount)=insrc_old(i)                  
        endif
      enddo
!      nsrc=nsmul*nsrc
       nsrc=nmulcount
    end if 




  end subroutine sou_clus

  !---------------------------------------------------------------------
  subroutine fill_arrs(xms,yms,nsrc,ecut2)
  !---------------------------------------------------------------------

    real, dimension(max_fit), intent(in) :: xms, yms
    real, intent(in) :: ecut2
    integer, intent(in) :: nsrc

    integer :: ierr,i,j,k,l,ipoint,ns,ne,ixs,iys,ixd,iyd,iamp, &
               x_start,x_end,y_start,y_end,ix,iy,i_offax,i_posang
    integer, dimension(max_ins) :: no_ev(max_ins)
    integer :: cutsum, masksum
    integer :: evmax, sum, ixmax, iymax, tempsum
    real    :: ss,ecut,xd,yd,xmax,xmin,ymax,ymin,offax,posang,x,y
    logical :: end_ev

    ierr=0

    xmax=xms(1)
    xmin=xms(1)
    ymax=yms(1)      
    ymin=yms(1)

    do k=1,nsrc
      xmax=max(xms(k),xmax)
      xmin=min(xms(k),xmin)
      ymax=max(yms(k),ymax)
      ymin=min(yms(k),ymin)        
    end do


    ecut = sqrt(ecut2)
    mask%ix_max=min(int(xmax+ecut),npix_arr(1,1))
    mask%ix_min=max(int(xmin-ecut),1)
    mask%iy_max=min(int(ymax+ecut),npix_arr(2,1))
    mask%iy_min=max(int(ymin-ecut),1)

! first check whether maskfrac criterion is met for each source
! (changed 05/01/07, GL):

    maskzero=0
      do k=1,nsrc
        do i=1,nim
          masksum=0
          cutsum=0
          do ix=mask%ix_min,mask%ix_max
            do iy=mask%iy_min,mask%iy_max
              if ((ix-xms(k))**2+(iy-yms(k))**2.le.ecut2) then
                cutsum=cutsum+1
                if (.not. detfla .or. mask_image(i,ixb(ix,i),iyb(iy,i))) then       !(JR)
                  if (.not. expfla .or. exp_image(ixb(ix,i),iyb(iy,i),i)>0.) then   !(JR)
                    masksum=masksum+1
                  endif
                endif
              endif
            end do
          end do
          if (cutsum > 0) maskfrac(k,i)=float(masksum)/float(cutsum)
!         do not fit source below threshold for maskfrac: 
          if (maskfrac(k,i) .lt. maskthreshold) maskzero(k,i)=1
!        write(*,*) "k,i,maskfrac(k,i),maskzero(k,i)",k,i,maskfrac(k,i),maskzero(k,i)
        end do
      end do 

    mask%mask(:,:,:)=0
    evmax=0
    do ix=mask%ix_min,mask%ix_max
      do iy=mask%iy_min,mask%iy_max
        mask%mask(:,ix-mask%ix_min+1,iy-mask%iy_min+1) = 0  
        do k=1,nsrc          
          if ((ix-xms(k))**2+(iy-yms(k))**2.le.ecut2) then
            sum=0
            do i=1,nim
              ! search brightest pixel   
              if (hotfla) then 
                sum=sum+image(ipoint,ixb(ix,ipoint),iyb(iy,ipoint)) !(JR)
                if (sum>evmax) then
                  evmax=sum
                  ixmax=ix
                  iymax=iy
                end if 
              end if
!              mask%mask(i,ix-mask%ix_min+1,iy-mask%iy_min+1) = 1
              if (detfla .or. expfla) then  
                if (maskzero(k,i) == 0) then
                  if (.not. detfla .or. mask_image(i,ixb(ix,i),iyb(iy,i))) then       !(JR)
                    if (.not. expfla .or. exp_image(ixb(ix,i),iyb(iy,i),i) > 0.) then !(JR)
                      mask%mask(i,ix-mask%ix_min+1,iy-mask%iy_min+1) = 1
                    endif
                  endif
                endif
              else
                mask%mask(i,ix-mask%ix_min+1,iy-mask%iy_min+1) = 1
              endif
            end do
          end if
        end do
      end do
    end do 
    if (hotfla) mask%mask(:,ixmax-mask%ix_min+1,iymax-mask%iy_min+1) = 0

!    write(*,*) "mask src1 :",int(xms(1)),int(yms(1)),mask%mask(6,int(xms(1))-mask%ix_min+1,int(yms(1))-mask%iy_min+1)
!    write(*,*) "mask src2 :",int(xms(2)),int(yms(2)),mask%mask(6,int(xms(2))-mask%ix_min+1,int(yms(2))-mask%iy_min+1)

    ! now check fraction of cutout radius cover by valid pixels,
    ! zero mask, if fraction < maskthreshold:
    !! shifted to beginning of fill_arrs (05/01/07)


!    maskzero=0
!    cutsum=0
!    masksum=0
!      do k=1,nsrc
!        do i=1,nim
!          masksum=0
!          cutsum=0
!          do ix=mask%ix_min,mask%ix_max
!            do iy=mask%iy_min,mask%iy_max
!              if ((ix-xms(k))**2+(iy-yms(k))**2.le.ecut2) then
!                cutsum=cutsum+1
!                masksum=masksum+mask%mask(i,ix-mask%ix_min+1,iy-mask%iy_min+1)
!                write(*,*) "mask:",i,ix,iy,mask%mask(i,ix-mask%ix_min+1,iy-mask%iy_min+1)
!              endif
!            end do
!          end do
!          if (cutsum > 0) maskfrac(k,i)=float(masksum)/float(cutsum)
!!         do not fit source below threshold for maskfrac: 
!          write(*,*) "k,i,masksum,cutsum",k,i,masksum,cutsum
!          if (maskfrac(k,i) .lt. maskthreshold) then
!            maskzero(k,i)=1
!            if (detfla .or. expfla) then
!              do ix=mask%ix_min,mask%ix_max
!                do iy=mask%iy_min,mask%iy_max
!                  mask%mask(i,ix-mask%ix_min+1,iy-mask%iy_min+1) = 0
!                enddo
!             enddo
!            endif
!          endif
!        write(*,*) "k,masksum,cutsum",k,masksum,cutsum
!        end do
!      end do    

!     write(*,*) "maskfrac 1:",maskfrac(1,:)
!     write(*,*) "maskfrac 2:",maskfrac(2,:)

    l = 1
    do i=1,num_instr
        
      if (num_evlist(i).eq.0) then

        do j=1,num_image(i) 
          ipoint=ip(i,j)
          do iy=mask%iy_min,mask%iy_max      
            do ix=mask%ix_min,mask%ix_max
              if (mask%mask(ipoint,ix-mask%ix_min+1, &
                            iy-mask%iy_min+1).eq.1) then 
                event(l)%n=image(ipoint,ixb(ix,ipoint),iyb(iy,ipoint))
                event(l)%ipt=ipoint
                l=l+1
              end if 
            end do
          end do
        end do

      else

! read in event lists
        ns=ip(i,1)
        ne=ip(i,num_image(i))

! open events file and get list of small maps
!cc       call pix2sky(float(ix_min)
!cc       call pix2sky(
        call get_smaps (no_ev(i),x_start,y_start,x_end,y_end)

! go through list of small maps and read one event at a time
        do while (.not.end_ev) 
          call get_event (no_ev(i),ixs,iys,ixd,iyd,iamp,end_ev)
          call sky2pix(ixs,iys,x,y,start_arr(1,i),step_arr(1,i))
          ix=nint(x)
          iy=nint(y)
          if (ix.ge.mask%ix_min.and.ix.le.mask%ix_max.and. &
              iy.ge.mask%iy_max.and.iy.le.mask%iy_max.and. &
                mask%mask(1,ix-mask%ix_min+1,iy-mask%iy_min+1).eq.1) &
          then
            event(l)%xs  = x
            event(l)%ys  = y 
            event(l)%n   = 1
            event(l)%ins = i
            do j=ns,ne
              if (iamp.ge.ampl_low(j).and.iamp.le.ampl_high(j)) &
                event(l)%ipt = j
            end do 
            xd=ixd-xdcen(i)
            yd=iyd-ydcen(i)
            offax = sqrt(xd*xd+yd*yd)*detpix(i)/60.
            i_offax = min(int(offax/epsmax(i)*max_offax)+1,max_offax)
            event(l)%offax  = i_offax
            posang = atan2(xd,yd)
            i_posang = min(int(posang/twopi*max_posang)+1,max_posang)
            event(l)%posang = i_posang
            hist(i)%hist(i_offax,i_posang)=hist(i)%hist(i_offax,i_posang)+1.
            l = l+1 
          end if
        end do

! accumulate off axis & position angle histogram        
        hist(i)%ns_offax=max_offax  
        hist(i)%ne_offax=1  
        hist(i)%ns_posang=max_posang  
        hist(i)%ne_posang=1  

        ss=0.
        do i_offax=1,max_offax
          do i_posang=1,max_posang
            if (hist(i)%hist(i_offax,i_posang).gt.0.) then
              hist(i)%ns_offax=min(hist(i)%ns_offax,i_offax)
              hist(i)%ne_offax=max(hist(i)%ns_offax,i_offax)
              hist(i)%ns_posang=min(hist(i)%ns_posang,i_posang)
              hist(i)%ne_posang=max(hist(i)%ns_posang,i_posang)
              ss=ss+hist(i)%hist(i_offax,i_posang)
            end if
          end do
        end do

        hist(i)%hist(:,:) = hist(i)%hist(:,:)/ss       

      end if       

    end do

    npts = l-1

  end subroutine fill_arrs

  !---------------------------------------------------------------------
  subroutine sky2pix (xsky,ysky,x,y,start,step)
  !---------------------------------------------------------------------

    integer          xsky,ysky
    real             x,y
    double precision start(2),step(2)

    x=(xsky-start(1)+0.5)/step(1)
    y=(ysky-start(2)+0.5)/step(2)

  end subroutine sky2pix

  !---------------------------------------------------------------------
  subroutine pix2sky(x,y,xsky,ysky,start,step)
  !---------------------------------------------------------------------

    real              x,y
    integer           xsky,ysky
    double precision  start(1),step(1)

    xsky=nint(x)
    ysky=nint(y)

  end subroutine pix2sky

  !---------------------------------------------------------------------
  subroutine rd_data (ierr)
  !---------------------------------------------------------------------
 
    integer           i,j,ipoint,maxdim,ierr
    parameter         (maxdim=2)
    integer           dattyp,filtyp,iomode, ipxid, ix, iy
    integer, dimension(max_ins) :: num_ima0
    real              exposure      
    real(kind=single), dimension(3) :: xnull, ynull
    double precision  start(maxdim),step(maxdim)
    character*1024    fname

    dattyp   = 10
    filtyp   = 1
    iomode   = 0
    xnull = 0.
    ynull = 0.
    xid_arr=-1
    crpix1=-2d30
    crpix2=-2d30
    crval1=-2d30
    crval2=-2d30
    crdelt1=-2d30
    crdelt2=-2d30

    nim = ip(num_instr,num_image(num_instr))

! add XID  band to pointer array ip if at least band 1-3 are present
    num_ima0 = num_image
    do i = 1,num_instr
      j=expip(i)
      if (num_image(i) >= xiddef(2,j) .and. xidfla) then
        num_image(i)=num_image(i)+1
        nim=nim+1
        ip(i,num_image(i)) = nim
        xid_arr(i)=num_image(i)
        pimin(nim)=pimin(ip(i,xiddef(1,j)))
        pimax(nim)=pimax(ip(i,xiddef(2,j)))
      end if
    end do

    do i = 1,num_instr
      do j = 1,num_ima0(i)
        ipoint=ip(i,j)  

        fname = c_image(ip(i,j))

        call co_image(fname,start,step,npix,exposure,xcen(i),ycen(i), &
                      ra_pnt(i),dec_pnt(i),pa_pnt(i),datetime, &
                      crpix1,crpix2,crval1,crval2,crdelt1,crdelt2, &
                      crota2,xnull,ynull,expip(i),ipoint,nim)

        pixim = 3600.0 * dabs(crdelt1) 

        if (j.eq.1) then
          start_arr(:,i) = start
          step_arr (:,i) = step 
          npix_arr (:,i) = npix 
        end if 

        fname = c_bgima(ip(i,j))

        call co_bgima(fname,npix,crpix1,crpix2,crval1,crval2,crdelt1,crdelt2, &
                      xnull,ynull,expip(i),ipoint,nim)

        if (expfla) then
           
          fname = c_expima(ip(i,j))
          call co_expima(fname,npix,crpix1,crpix2,crval1,crval2,crdelt1,crdelt2, &
                         xnull,ynull,expip(i),ipoint, &
                         xcen(i),ycen(i),exp_center(ipoint),nim)

          !exp_center(ipoint)=exp_image(int(xcen(i)),int(ycen(i)),ipoint)
        else
          if (i==1.and.j==1) allocate (exp_image(imbuff,imbuff,nim))  !(JR)
          exp_image=1.0          
        end if

      end do

      if (detfla) then
        fname = c_detima(i)
        call co_mask(fname,npix,crpix1,crpix2,crval1,crval2,crdelt1,crdelt2, &
                     xnull,ynull,expip(i),ip(i,1),nim)
        
        do j = 2,num_image(i)
          do iy = 1,npix(2)
            do ix = 1,npix(1)
             mask_image(ip(i,j),ixb(ix,ip(i,j)),iyb(iy,ip(i,j)))= &   !(JR)
                mask_image(ip(i,1),ixb(ix,ip(i,1)),iyb(iy,ip(i,1)))   !(JR)
            end do
          end do
        end do
      else
        if (i==1) allocate (mask_image(nim,imbuff,imbuff))            !(JR)
        mask_image=.true.
      end if   
        
    end do

    if (crpix1<-1d30) crpix1=0d0
    if (crpix2<-1d30) crpix2=0d0
    if (crval1<-1d30) crval1=0d0
    if (crval2<-1d30) crval2=0d0
    if (crdelt1<-1d30) crdelt1=0d0
    if (crdelt2<-1d30) crdelt2=0d0

! add XID band images if band 1 - 3 are available

    do i = 1,num_instr
     j=expip(i)
     if (num_image(i) >= xiddef(2,j) .and. xidfla) then 
      ipxid=ip(i,num_image(i))
      do iy=1,npix(2)
        do ix=1,npix(1)
!-------------------------------------------------------------------------------------------------
!--------------- (JR) ----------------------------------------------------------------------------
          image(ipxid,ixb(ix,ipxid),iyb(iy,ipxid)) = &
            image(ip(i,xiddef(1,j)),ixb(ix,ip(i,xiddef(1,j))),iyb(iy,ip(i,xiddef(1,j)))) + &
            image(ip(i,xiddef(2,j)),ixb(ix,ip(i,xiddef(2,j))),iyb(iy,ip(i,xiddef(2,j))))
          bg_image(ixb(ix,ipxid),iyb(iy,ipxid),ipxid) = &
            bg_image(ixb(ix,ip(i,xiddef(1,j))),iyb(iy,ip(i,xiddef(1,j))),ip(i,xiddef(1,j))) + &
            bg_image(ixb(ix,ip(i,xiddef(2,j))),iyb(iy,ip(i,xiddef(2,j))),ip(i,xiddef(2,j)))
          exp_image(ixb(ix,ipxid),iyb(iy,ipxid),ipxid) = &
            (exp_image(ixb(ix,ip(i,xiddef(1,j))),iyb(iy,ip(i,xiddef(1,j))),ip(i,xiddef(1,j))) + &
             exp_image(ixb(ix,ip(i,xiddef(2,j))),iyb(iy,ip(i,xiddef(2,j))),ip(i,xiddef(2,j)))) / 2.
          mask_image(ipxid,ixb(ix,ipxid),iyb(iy,ipxid)) = &
            mask_image(ip(i,xiddef(1,j)),ixb(ix,ip(i,xiddef(1,j))),iyb(iy,ip(i,xiddef(1,j)))) .and. &
            mask_image(ip(i,xiddef(2,j)),ixb(ix,ip(i,xiddef(2,j))),iyb(iy,ip(i,xiddef(2,j))))
!-------------------------------------------------------------------------------------------------
!-------------------------------------------------------------------------------------------------
        end do
      end do
      exp_center(ipxid)=(exp_center(ip(i,xiddef(1,j)))+ &
                         exp_center(ip(i,xiddef(2,j))))/2.0
      energy(ipxid)=(energy(ip(i,xiddef(1,j)))+energy(ip(i,xiddef(2,j))))/2.0 
      ecf(ipxid)=xidecf(i)
     end if
    end do

    allocate (bg_image0(imbuff,imbuff,nim))
    bg_image0 = bg_image       
 
  end subroutine rd_data

  !---------------------------------------------------------------------
  subroutine sorind(a,w,ind,indw,n,nw,ierr)
  !---------------------------------------------------------------------

    integer :: n,nw,ierr,ind(n),indw(nw),i,j
    real :: a(n),w(nw)

    ierr=0

! index the array ind such that a is in ascending order
    do i = 1,n
      ind(i) = 0
    enddo
    do i = 1,nw
      indw(i) = 0
    enddo

    if (n.gt.1) then
      call indexr(n,a,ind)
    else if (n.eq.1) then
      ind(1) = 1
    else if (n.eq.0) then
      ind(1) = 0
    endif

! sort a according to ind
    do i = 1,n
      w(i) = a(ind(i))
    enddo
    do j = 1,n
      a(j) = w(j)
    enddo
    do j = 1,n
      indw(j) = ind(j)
    enddo

  end subroutine sorind    


  !---------------------------------------------------------------------
  subroutine emldetect
  !---------------------------------------------------------------------

!  author : GRH              date: 15-APR-1987 15:48
!  update : GRH              date: 3-DEC-1992 09:13  
!  update : PTK              date: 05-JAN-1993 17:27 EXSAS implementation
!  update : PTK              date: 14-JUN-1993 20:17 table output / correct
!                                                    bugs
!  update : EXSAS            date: 15-OCT-1993 08:52
!  update : PTK              date: 08-NOV-1993 10:41 POSFLA, new parameter
!                                                    IPOS in CGAUSN
!  update : PTK              date: 18-JAN-1994 09:55 NELEM1,NELEM2
!  update : EXSAS            date: 19-JAN-1994 16:43
!  update : HEB              date: 27-MAY-1996

!     character*80  table
!     character*2   io
!     character*70  d2par

      integer ::    k, ll, ierr, &
                    indx(max_sou),insrc(max_fit), &
                    iwork(max_sou),kk,nsou, &
                    in(max_sou), nsrc, istat, id_clu 
      real ::       scut2, ecut2, xx(max_sou), yy(max_sou), extent(max_sou)
      real ::       totcounts(max_sou),totrate(max_sou)
      real(kind=single), dimension(max_fit) :: xms, yms, exms, eyms, eradec 
      real(kind=single), dimension(max_ima, max_fit) :: scts, escts, rate, erate
      real(kind=single), dimension(max_ima, max_fit) :: detcounts,srcexpos,vignetting,srcontime
      real(kind=single), dimension(max_ima, max_sou) :: src

      real(kind=single), dimension(max_sou) :: sn, rwork
      real(kind=single), dimension(max_sou) :: selvalue
      real(kind=double), dimension(max_fit) :: ra, dec, lii, bii
      integer, dimension(max_ins,max_fit) :: chipix,chipiy,ccdinr,istate1 
      real(kind=double), dimension(max_ins,max_fit) :: offax
 
      real, dimension(max_ima,max_fit) :: exi_ml 
      real, dimension(max_fit) :: ext, ext_err, ext_ml 
      real, dimension (max_ima) :: enfrac
      character(len=5) :: flags
      logical :: first, fitdone                    
      integer :: nsmul_old
      call mulmlpari

! Read input images
      call rd_data (ierr)

      call get_sources (nsou,xx,yy,sn,src,extent,in,totrate,totcounts)

      
! set threshold for nmulsou fitting:
      if (threshfla) then
         if (threshcol == "LIKE" .or. threshcol == "like") selvalue=sn
         if (threshcol == "RATE" .or. threshcol == "rate") selvalue=totrate
         if (threshcol == "SCTS" .or. threshcol == "scts") selvalue=totcounts
      endif      

! Sort the array, key is SN
      call sorind(sn, rwork, indx, iwork, nsou, max_sou, istat)


      iwork=0
      id_clu=0
      ll=0
       
! Do loop, start from brightest source 
      nsmul_old=nsmul 
      do k=1,nsou
        fitdone = .false.
!             write(*,*) k, " sources processed"
        if (mod(k,10) == 0) then
           write(errstr,*) k, " sources processed"
           call message(AppMsg,Verbose)
        end if
        kk=indx(nsou-k+1)
        if (iwork(kk).eq.0) then
          iwork(kk)=1
          id_clu=id_clu+1





! Calculate cut radii
          call cut_radius (xx(kk),yy(kk),scut2,ecut2,enfrac)
         ! initialize nsrc and ext_ml(1): (3/8/05)
         nsrc=1
         ext_ml(1)=0.0

         if (twostage) then
           first=.true.
           nsmul=1

           ! Identify source cluster
           call sou_clus (xms,yms,scts,ext,insrc,nsrc,scut2,xx,yy,src,extent,in, &
                          k,nsou,indx,iwork,threshold,selvalue) 

           ! Fill data arrays
           call fill_arrs (xms,yms,nsrc,ecut2)
           ! change "exit" to  "if then" (5/5/06, GL)
           if (npts.gt.((nim+3)*nsrc+1)) then
             ! Perform multi-source ML fit
              call mulsou(nsrc,xms,yms,ext,scts,rate,ra,dec,lii,bii, &
                          chipix,chipiy,ccdinr,offax, &
                          exms,eyms,ext_err,escts,erate,eradec, &
                          exi_ml,ext_ml,ecut2,srcexpos,vignetting,srcontime,insrc,first)
              fitdone = .true.
           endif
         endif

         if (.not. twostage .or. (maxval(ext_ml(1:nsrc)) .gt. 0)) then
            nsmul=nsmul_old
            first=.false.
            call sou_clus (xms,yms,scts,ext,insrc,nsrc,scut2,xx,yy,src,extent,in, &
                           k,nsou,indx,iwork,threshold,selvalue) 
            call fill_arrs (xms,yms,nsrc,ecut2)
           ! change "exit" to  "if then" (5/5/06, GL)
           if (npts.gt.((nim+3)*nsrc+1)) then
              call mulsou(nsrc,xms,yms,ext,scts,rate,ra,dec,lii,bii, &
                          chipix,chipiy,ccdinr,offax, &
                          exms,eyms,ext_err,escts,erate,eradec, &
                          exi_ml,ext_ml,ecut2,srcexpos,vignetting,srcontime,insrc,first)
              fitdone = .true.
           endif
         endif



! Set dummy output values
            flags="     "
           
! write source info to table

!            write(*,*) "wr_stab:", xms(1),yms(1),ext(1),rate(1:10,1)
!            if (nsrc > 1) write(*,*) "wr_stab:", xms(2),yms(2),ext(2),rate(1:10,2)
          if (fitdone) then
            call wr_stab(ll,nsrc,insrc,id_clu,xms,yms,scts,rate,ra,dec,lii,bii,&
                         chipix,chipiy,ccdinr,offax, &
                         exms,eyms,escts,erate,eradec,exi_ml,ext,ext_err,&
                         ext_ml,enfrac,ecut2,flags,srcexpos,vignetting,srcontime)
          endif
        endif

      end do

! Close output table       
      ! ll = -ll
      call mlout(0,ll,k,0,0,0,0,0.,0.,0.,0.,0., &
                 0.,0.,0.,&
                 0.,0.,0.,0.,0.,0.,0.,0., &
                 0d0,0d0,0.,0d0,0d0,0,0,0,0.,0.,0.,0.,0.,0.,0.,0.,0., &
                 0.,0.,0.,0.,0.,0.,flags) 


       

! write out source map
      if (smpfla) call wr_srcmap

    end subroutine emldetect
  
end module emldetect_mod








