module fftpsf

  use param
  use dal
  use cal
  use emldata
  use edetect_io_ml
  
  implicit none

  private


  public :: fpsf


  contains

!---------------------------------------------------------------------------------------

  subroutine fpsf(energy,iinst,iim,xremain,yremain,x0,y0,sig,ecut,isize,ns)

    real,    intent(in) :: ecut
    real(kind=double), intent(in) ::  x0, y0, sig, xremain, yremain
    integer, intent(in) :: iinst,iim,ns
    integer             :: ipoint,iband   !!!,nbands,nsold=-1
!    real,    intent(in) :: ene
    real, dimension(max_ima), intent(in) :: energy
    integer, intent(out) :: isize
    real ::  modval
    real(kind=double), dimension(max_fit) :: ra,dec

    integer, dimension(max_mask,max_mask) :: xsort,ysort
    real, dimension(:,:,:), allocatable        :: sima
    real, dimension(:,:,:,:), allocatable        :: bima0
    real, dimension(:,:),  allocatable        ::  bima,mima
    real, dimension(:,:,:),  allocatable        ::  mima0

!
!!!!    real, dimension(:,:), allocatable, public :: fima
!       defined in emldata.f90
      
    complex, dimension(:,:),  allocatable     ::  ffmodel,ffpsf,prodima
!    integer(kind=int32)                                    :: pplan,mplan,fplan
    integer*8                                    :: pplan,mplan,fplan
    real, dimension(:,:), pointer :: pima

    type (PsfT)        :: ppsf
    type(PsfBinSizeT), dimension(max_ima)  :: bsiz
    type(PSFwindowT)   :: window
    type(PsfValidityRangesT) :: vrange 
    real(kind=double) :: x,y,x1,y1
!!!    real(kind=double) :: dx,dy,ddx,ddy,xd0,yd0,xx,yy
    real(kind=double) :: dx,dy,ddx,ddy,xx,yy
    real(kind=double) , dimension(1:max_fit) :: x0old,y0old,sigold
!!!    real :: ssum,winr,pfacX,pfacY,xyoff2,ecutold,fluxfac,fftscale,sig4sq
    real :: ssum,winr,pfacX,pfacY,fluxfac,fftscale,sig4sq
    real :: sigmin=0.1,norm

!!! IT 2012/01
!    integer             :: i,j,ix,iy,ixx,iyy,lx,ly,kx,ky,kk,np,np2,isig3,isig3sq
    integer             :: i,j,ix,iy,np
    real                :: np2
    real(kind=double)   :: pxmin,pxmax,pymin,pymax,dx1,dx2,dy1,dy2
    integer             :: pxmin0,pymin0,pxmax0,pymax0,pxmin1,pymin1
!!!---------------
    integer             :: isx,isy,isizeold=-1,isizehalf,n,iecut,ixmin,ixmax,iymin,iymax,ninst
    integer, dimension(max_ima) :: nx, ny
    integer, dimension(max_ins,max_fit)  :: chipix, chipiy, ccdinr 
    real(kind=double), dimension(max_ins,max_fit) :: offax, azim
!!! IT 2012/01
!---JR
!    real :: testsum
!    integer       ix1, iy1
!!!---------------


!!! IT 2012/01
!!! sec2mm will be calculated for each instrument individually, using the CAL
!    real, parameter :: sec2mm = 7500.0*pi/(180.0*3600.0)
    real(kind=double), dimension(max_ins) :: sec2mm
    real(kind=double) :: E, theta, phi, val, focalLength
!    real(kind=double) :: E, theta, phi, val
!!!---------------
    logical :: first = .true., allocated=.false.
    logical :: havePSF 
!    real :: t1,t2
!    integer :: clock1,clock2,clock3

!    integer :: istate1

    
      INTEGER, PARAMETER :: FFTW_MEASURE=0
      INTEGER, PARAMETER :: FFTW_EXHAUSTIVE=8
      INTEGER, PARAMETER :: FFTW_PATIENT=32
      INTEGER, PARAMETER :: FFTW_ESTIMATE=64
      INTEGER, PARAMETER :: FFTW_ESTIMATE_PATIENT=128

    save

!    write(*,*) "start fpsf with ns,sig,iinst:",ns,sig,iinst


!!! IT 2012/01: not used
!    nbands=num_image(iinst)
!!!---------------


!    write(*,*) "ns,sig:",ns,sig

    count_psf=count_psf+1

    if (first) then
      sigold(:)=-1.0
      y0old(:)=-1.E9
      x0old(:)=-1E9
      havePSF=.false.
      allocate( sima(0:650,0:650,nim) ) !! increased to 450 - JR
!      write(*,*) "allocate sima"
    endif 

!    determine subimage array sizes (use 2**n sizes for FFT)  
    
     do n=1,8
       isize=2**n
       if (nint(2*ecut)+2 .lt. isize .or. isize .ge. max_mask) exit
     end do
    isizehalf=isize/2

     iecut=int(ecut)+1
!     write(*,*) "iecut:",iecut
     
!   write(*,*) "isize=",isize

   if (isizeold /= isize) then
!     write(*,*) "isize:",isize 
     havePSF=.false.
     fftscale=1./float((isize*isize))
     if (.not. first) then
       if (allocated) then
         deallocate(mima)
         deallocate(mima0)
         deallocate(ffmodel)
         deallocate(ffpsf)
         deallocate(prodima)
         allocated = .false.
       endif
       deallocate(bima0)
       deallocate(bima)
       deallocate(fima)
     endif
!       write(*,*) "allocate fft images"
     if (.not. allocated) then     
       allocate( mima0(1:isize,1:isize,max_fit) )
       allocate( mima(1:isize,1:isize))
       allocate( ffmodel(1:isizehalf+1,1:isize) )
       allocate( ffpsf(1:isizehalf+1,1:isize) )
       allocate( prodima(1:isizehalf+1,1:isize) )
       allocated = .true.
     endif
     allocate( bima(1:isize,1:isize) )
     allocate( bima0(1:isize,1:isize,nim,max_fit) )
     allocate( fima(1:isize,1:isize) )
!      create sorting arrays to rearrange quadrants of kernel image for FFT
!
!    1|2         4|3
!    ---         ---
!    3|4    =>   2|1 
!
       do i=1,isize
        do j=1,isize
          if (i .le. isizehalf) then
            xsort(i,j) = isizehalf+i
          else
            xsort(i,j) = i-isizehalf
          endif
          if (j .le. isizehalf) then
            ysort(i,j) = isizehalf+j
          else
            ysort(i,j) = j-isizehalf
          endif             
        enddo
      enddo 
      isizeold=isize 

      ! create FFTW plans:
      if (extfla  .or. input_ext) then
!        write(*,*) "make plans",isize,ecut
        call sfftw_plan_dft_r2c_2d(pplan,isize,isize,bima,ffpsf,FFTW_MEASURE)
        call sfftw_plan_dft_r2c_2d(mplan,isize,isize,mima,ffmodel,FFTW_MEASURE)
        call sfftw_plan_dft_c2r_2d(fplan,isize,isize,prodima,fima,FFTW_MEASURE)
      endif
    
   endif

     



    if (newcal) then 
      havePSF=.false.
    endif
    

!--- (JR) ------------------------------------------------------------------------
    if (psfmod == "ellbeta") then
        modpsf = .true.
    else
        modpsf = .false.
    endif
!---------------------------------------------------------------------------------


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! 1st ninst-iband loop: copy PSF from CAL and prepare for 2nd loop
!!! only to be done at the first call ("havePSF" false)


   if (.not. havePSF) then 

!   use subroutine rawcoords to calculate theta and phi:
    call impix2equ(sngl(x0),sngl(y0),ra(1),dec(1))
    call rawcoords(ra,dec,1,datetime,chipix,chipiy,ccdinr,offax,azim)
    

! Note on azimuth angles: The CAL input/output angle phi projects clockwise onto 
!  sky images. PA_PNT is measured anti-clockwise from the north direction.
! phi as needed as input for the CAL PSF here is measured clockwise from the East direction. 

    np = max(0,nint((pixim-1.0)*0.5))
    np2= 1./(2*np+1)**2

    do ninst=1,num_instr
      theta =  offax(ninst,1)
      phi = azim(ninst,1)-pa_pnt(ninst)*pi180+pi2
      if (phi .gt. twopi) phi=phi-twopi      
      if (phi .lt. -twopi) phi=phi+twopi  
!----- azi -----------------------------------
!      azim(ninst,1)=azim(ninst,1)+2.*pi-pa_pnt(ninst)/180.*pi
!      if (azim(ninst,1) .gt. 2*pi) azim(ninst,1)=azim(ninst,1)-2.*pi
!      if (azim(ninst,1) .lt. -2.*pi) azim(ninst,1)=azim(ninst,1)+2.*pi
!---------------------------------------------
!      write(*,*) "fpsf:x,y,theta,phi:",x0,y0,theta,azim(ninst,1),pa_pnt(ninst),phi
!---------------------------------------------------------------------------------


!---------------------------------------------------------------------------------

!!! IT 2012/01

!!! replaced "expip(iinst)" by "expip(ninst)": we need to refer to the
!!! instrument number which is given by the loop counter, not the the one which
!!! emlfit uses when calling fpsf.

      if (extpsf) then
        if (expip(ninst) .eq. 1) call CAL_setState1(instrument=EPN, &
                                            AccuracyLevel=ACCURACY_EXTENDED) 
        if (expip(ninst) .eq. 2) call CAL_setState1(instrument=EMOS1, &
                                            AccuracyLevel=ACCURACY_EXTENDED)
        if (expip(ninst) .eq. 3) call CAL_setState1(instrument=EMOS2, &
                                            AccuracyLevel=ACCURACY_EXTENDED)
      else
!----------------------------------------------------------
!----- JR -------------------------------------------------
!--PN
      if (expip(ninst) .eq. 1) then
        if (modpsf) then
        call CAL_setState1(instrument=EPN, &
                          AccuracyLevel=ACCURACY_ELLBETA)
          else
          call CAL_setState1(instrument=EPN, &
                          AccuracyLevel=ACCURACY_MEDIUM)
        endif
      endif
!--MOS1
      if (expip(ninst) .eq. 2) then
        if (modpsf) then
        call CAL_setState1(instrument=EMOS1, &
                          AccuracyLevel=ACCURACY_ELLBETA)
          else
          call CAL_setState1(instrument=EMOS1, &
                          AccuracyLevel=ACCURACY_MEDIUM)
        endif
      endif
!--MOS2
      if (expip(ninst) .eq. 3) then
        if (modpsf) then
        call CAL_setState1(instrument=EMOS2, &
                          AccuracyLevel=ACCURACY_ELLBETA)
          else
          call CAL_setState1(instrument=EMOS2, &
                          AccuracyLevel=ACCURACY_MEDIUM)
        endif
      endif
!----------------------------------------------------------
!----------------------------------------------------------
        psfnorm=1.0
      endif



!!! IT 2012/01
!!! get focal length of the telescope from the CAL (for each instrument)
!!! calculate conversion factors "sec2mm" (arcsec to mm)
      if (first) then
           !!! IT 2012/03: new version only for ellbeta to be consistent
           !!! with old medium-accuracy positions
           if (modpsf) then
              call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLength)
              sec2mm(ninst) = focalLength*pi180/3600.d0
           else
              sec2mm(ninst) = 75d0*pi180/36d0    ! 7500d0*pi/(180.0*3600.0)
           endif
      endif


!!!!!!!!!!
!        xd0=x0-xcen(ninst)
!        yd0=y0-ycen(ninst)
!        theta=sqrt(xd0**2+yd0**2)*pixim
!        phi=atan2(xd0,yd0) + pi/2.0
!!!!!!!!!!!

      vrange = CAL_PsfValidityRanges()
      theta = max(min(theta,vrange%thetaMax-1.0),vrange%thetaMin+1.0)

!!      np = max(0,nint((pixim-1.0)/2.0))
!!      np2= 1./(2*np+1)**2
!!
      do iband=1,num_image(ninst)

      winr = int(ecut)+2.0

!-------------------------------------
!-- JR -------------------------------
      if (modpsf) then
      winr=  min(winr*pixim,400.0)*sec2mm(ninst)
!---------------------------------------------------------
        else
      winr=  min(winr*pixim,100.0)*sec2mm(ninst)
      endif
!-------------------------------------
!-------------------------------------
      window%halfX = winr
      window%halfY = winr

        ipoint=ip(ninst,iband)
!        write(*,*) "iband:",iband
        E = energy(ipoint)
!        write(*,*) "Energy:",E
        E = max(min(E,vrange%eMax-1.0),vrange%eMin+1.0)



!!-------------------------------------
!!-- JR -------------------------------
!      if (modpsf) then
!!        azim(ninst,1)=0.0
!!        pa_pnt(ninst)=0.0
!        ppsf = Cal_getPSF(E, theta, azim(ninst,1))
!! -- Necessary for Richard's CAL -------------------
!        call CAL_psfSetPosAngle(ppsf, pa_pnt(ninst)/180.*pi)
!!---------------------------------------------------
!        else
!        ppsf = Cal_getPSF(E, theta, azim(ninst,1))
!        call CAL_psfSetPosAngle(ppsf, pa_pnt(ninst)/180.*pi)
!      endif
!-------------------------------------
      ppsf = Cal_getPSF(E, theta, azim(ninst,1))
      call CAL_psfSetPosAngle(ppsf, pa_pnt(ninst)*pi180)
!-------------------------------------
        bsiz(iband) = CAL_psfBINSIZE(ppsf)
!        write(*,*) "PSF binsize:", bsiz(iband)

        call CAL_psfImage (ppsf, window, pima) 

!       fudge factor for EXTENDED PSF:
!!! IT 2012/01: psfnorm is 1.
!!!        pima=pima*psfnorm
        
        nx(iband) = size(pima,1)
        ny(iband) = size(pima,2)

!!!-----------------------------------------------------------------------------
!!! IT 2012/01
!!! New re-sampling algorithm for the ellbeta 2d PSF. The PSF is re-binned
!!! within the next ninst-niband loop pixel by pixel, which means at this
!!! stage: just copy the psf array to sima.
!!! medium accuracy PSF: smoothing as before.

         if (modpsf) then

!!! CAUTION!!! pima indices start with 0, not with 1
           sima(0:nx(iband)-1,0:ny(iband)-1,ipoint) = pima

!!!------> old version of re-sampling (1)
!!!        smoothing the medium accuracy PSF with a box of image pixel size
         else
            ! smooth PSF with (2*np+1) sized box:
            do isy=np,ny(iband)-1-np 
               do isx=np,nx(iband)-1-np
!!! IT 2012/01: faster
!                   ssum=0.0
!                  do ixx=isx-np,isx+np
!                     do iyy=isy-np,isy+np
!                        ssum=ssum+pima(ixx,iyy)
!                     end do
!                  end do
! !            write(*,*) "a)",iim,np,ny(iim),nx(iim)
! !            write(*,*) "a)",isx,isy
!                  sima(isx,isy,ipoint)=ssum/np2
                  sima(isx,isy,ipoint) = &
                       sum(pima(isx-np:isx+np,isy-np:isy+np)) * np2
!!!
               end do
            end do
         endif
!!!<--------

!!!-----------------------------------------------------------------------------
!
!!------------------------------------------------------------------------------------------
!!-- test JR -------------------------------------------------------------------------------
!    testsum=0.
!    do ix1=0,nx(iband)-1
!     do iy1=0,ny(iband)-1
!      testsum=testsum+pima(ix1,iy1)
!     end do
!    end do
!!
!    write(*,*) "Normalization for source <<fftpsf.f90>>:", iband , testsum
!!------------------------------------------------------------------------------------------
!!------------------------------------------------------------------------------------------
       call CAL_releaseMemory(pima)

!      open(unit=1,file='sima.dat',status='unknown')
!      write(1,*) sima(1:30,1:30,1,1)
!      close(1)

      enddo

      enddo
      havePSF=.true.
    endif



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! 2nd ninst-iband loop: rebin the PSF from its CAL coordinates to image pixels
!!! only to be done, if the image coordinates / reference pixel have changed


  ! interpolate PSF values to  image pixel coordinates:

    if (x0 .ne. x0old(ns) .or. y0 .ne. y0old(ns)) then


!!! IT 2012/01: later
!!-------------------------------------
!!-- JR -------------------------------
!      if (modpsf) then
!        fluxfac = (pixim)*(pixim)
!        else
!        fluxfac=(sec2mm(ninst)*pixim)*(sec2mm(ninst)*pixim)
!      endif
!!-------------------------------------
!!-------------------------------------



!!!-----------------------------------------------------------------------------
!!! IT 2012/01
!!! new re-sampling algorithm
!!! (also implemented in psf.f90)
!!!
!!! Aim: accurate flux conservation and pixel position.
!!! For each image pixel, counts from the corresponding PSF area are summed
!!! up. To achieve flux conservation, we have to look for PSF pixels totally
!!! covered by an image pixel and for PSF pixels partly covered by an image
!!! pixel, either in x direction or in y direction or both. Counts are then
!!! weighted with the covered pixel area.
!!!     :    :    :    :       
!!!     :|   :    : |  :   
!!! ....:|...:....:.|..:... <- pymax0+1
!!!  ---------------------  <- pymax         dashed: image pixel
!!! ....:|...:....:.|...... <- pymax0        dotted: psf pixel
!!!     :|   :    : |  :   
!!!     :|   :    : |  :      
!!! .....|..........|...... <- pymin1
!!!     :|   :    : |  :   
!!!  ---------------------  <- pymin
!!!     :|   :    : |  :   
!!! ....:|...:....:.|..:... <- pymin0
!!!     :|   :    : |  :   
!!!     :    :    :    :      
!!!      ^pxmin     ^pxmax
!!!     ^pxmin0   ^pxmax0
!!!
!!! px/ymin give the lower left corner of the image pixel as PSF coordinates.
!!! px/ymin0 give the number of the PSF pixel in which (pxmin,pymin) is.
!!! pfacX/Y are the conversion factors between image and PSF coordinates, 
!!!         i.e. the size of an image pixel expressed in PSF pixels.
!!! px/ymax = px/ymin + pfacX/Y are, thus, the upper right corner of the image
!!!         pixel, given as PSF coordinates.
!!! px/ymax0 give the number of the PSF pixel in which (pxmax,pymax) is.
!!!
!!! I.e. PSF pixel corners have areas (pxmin-pxmin0)*(pymin-pymin0),
!!!                                   (pxmax0+1-pxmax)*(pymin-pymin0), etc.
!!!      PSF pixel slices  have areas (pxmin-pxmin0) * 1, (pxmax0+1-pxmax) * 1,
!!!                                   1 * (pymin-pymin0), 1 * (pymax0+1-pymax).
!!!      Whole PSF pixels are found between px/ymin0+1 and pymax0.


      if (modpsf) then        !!! -> ELLBETA (2D) PSF

         do ninst=1,num_instr

!!! IT 2012/01 with the new re-binning, the normalization factor is 1
!!!         fluxfac = pixim*pixim

            do iband=1,num_image(ninst)
               ipoint=ip(ninst,iband)
               bima0(:,:,ipoint,ns)=0.0
               np = max(0,nint((pixim-1.0)*0.5))

!!! xremain, yremain: offset of the model source (= PSF) center from the nearest
!!!                   image pixel [centre], abs <= 0.5
!!! reference xx, yy: lower left corner of this image pixel, expressed in PSF
!!!                   coordinates
!!!                   i.e.: half the PSF size (nx,ny) minus the offset

!!! xx,yy: window%halfX,Y is _NOT_ half the number of psf pixels - that's nx,ny / 2!

!        pfac=pixim*sec2mm/bsiz(iband)%binX
         pfacX=pixim*sec2mm(ninst)/bsiz(iband)%binX
         pfacY=pixim*sec2mm(ninst)/bsiz(iband)%binY

!!! center of PSF (nx/2.,ny/2.) - xremain,yremain - half an image pixel
         xx = dble(nx(iband)*0.5) - (xremain + 0.5d0 )*pfacX
         yy = dble(ny(iband)*0.5) - (yremain + 0.5d0 )*pfacY

!!! ixmin, ixmax, iymin, iymax: array ranges around image centre
         iymin=max(0,     isizehalf-iecut, isizehalf-int((ny(iband)/2-np)/pfacY))
         iymax=min(isize, isizehalf+iecut, isizehalf+int((ny(iband)/2-np)/pfacY))
         ixmin=max(0,     isizehalf-iecut, isizehalf-int((nx(iband)/2-np)/pfacX))
         ixmax=min(isize, isizehalf+iecut, isizehalf+int((nx(iband)/2-np)/pfacX))

         do j=iymin,iymax
            !! loop variables describing image pixel edges in PSF coordinates
            pymin = yy + dble(j-isizehalf)*pfacY                 !! formerly: y
            pymin0 = int(pymin)
            pymin1 = pymin0 + 1
            pymax = pymin + pfacY
            pymax0 = int(pymax)
            !! cut to usable area
!!!TO DO!!!
            !! coefficients used in the re-binning term
            dy1 = pymin1 - pymin
            dy2 = pymax - pymax0

            do i=ixmin,ixmax
               !! loop variables describing image pixel edges in PSF coordinates
               pxmin = xx + dble(i-isizehalf)*pfacX                !! formerly: x
               pxmin0 = int(pxmin)
               pxmin1 = pxmin0 + 1
               pxmax = pxmin + pfacX
               pxmax0 = int(pxmax)
               !! cut to usable area
!!!TO DO!!!
               !! coefficients used in the re-binning term
               dx1 = pxmin1 - pxmin
               dx2 = pxmax - pxmax0

               !! only, if complete pixels in x and / or y are covered:
               ssum=0.0
               if (pxmax0 > pxmin1 ) then
                  !! edges: complete in x, fraction in y
                  ssum = ssum +                                           &
                       dy1 * sum(sima(pxmin1:pxmax0-1, pymin0, ipoint)) + &
                       dy2 * sum(sima(pxmin1:pxmax0-1, pymax0, ipoint))
               endif
               if (pymax0 > pymin1 ) then
                  !! edges: complete in y, fraction in x
                  ssum = ssum +                                           &
                       dx1 * sum(sima(pxmin0, pymin1:pymax0-1, ipoint)) + &
                       dx2 * sum(sima(pxmax0, pymin1:pymax0-1, ipoint))
                  if (pxmax0 > pxmin1 ) then
                     !! whole (PSF) pixels
                     ssum = ssum +                                        &
                          sum(sima(pxmin1:pxmax0-1, pymin1:pymax0-1, ipoint))
                  endif
               endif

               !! always: corners, subpixel contributions
               ! i+1, j+1:
               ! indices i,j are centered around 0; bima indices start with 1
               bima0(i+1,j+1,ipoint,ns) = ssum +                          &
                    dx1 * dy1 * sima(pxmin0, pymin0, ipoint) +            &
                    dx1 * dy2 * sima(pxmin0, pymax0, ipoint) +            &
                    dx2 * dy1 * sima(pxmax0, pymin0, ipoint) +            &
                    dx2 * dy2 * sima(pxmax0, pymax0, ipoint)

          enddo                    !!! end "i=ixmin,ixmax"
         enddo                     !!! end "j=iymin,iymax"
!!!-----------------------------------------------------------------------------
         enddo                     !!! end "iband=1,num_image(ninst)"
         enddo                     !!! end "ninst=1,num_instr"

      else                    !!! -> MEDIUM ACCURACY PSF

         do ninst=1,num_instr

            fluxfac=(sec2mm(ninst)*pixim)*(sec2mm(ninst)*pixim)

            do iband=1,num_image(ninst)
               ipoint=ip(ninst,iband)
               bima0(:,:,ipoint,ns)=0.0
               np = max(0,nint((pixim-1.0)*0.5))

!        pfac=pixim*sec2mm/bsiz(iband)%binX
         pfacX=pixim*sec2mm(ninst)/bsiz(iband)%binX
         pfacY=pixim*sec2mm(ninst)/bsiz(iband)%binY

!!!------> old version of re-sampling (2)
!!!        interpolating the medium accuracy PSF from CAL px to image px
        xx=(window%halfX-pixim*sec2mm(ninst)*xremain)/bsiz(iband)%binX-0.5
        yy=(window%halfY-pixim*sec2mm(ninst)*yremain)/bsiz(iband)%binY-0.5

        iymin=max(1,isizehalf-iecut+1,isizehalf-int((ny(iband)/2-np)/pfacY)+1)
        iymax=min(isize,isizehalf+iecut+1,isizehalf+int((ny(iband)/2-np)/pfacY)+1)
        ixmin=max(1,isizehalf-iecut+1,isizehalf-int((nx(iband)/2-np)/pfacX)+1)
        ixmax=min(isize,isizehalf+iecut+1,isizehalf+int((nx(iband)/2-np)/pfacX)+1)


!        do j=1,isize
!        do j=isize/2-iecut+1,isize/2+iecut+1
         do j=iymin,iymax
!          23/06/06 (GL)
           y1=dble(j-isizehalf)-1.0
!          y1=float(j-isize/2)-1.0


          y=yy+y1*pfacY
          iy = min(max(int(y),np),ny(iband)-2-np)
          dy=max(y-iy,0.D0)
          ddy=max(1.D0-dy,0.D0) 
!          do i=isize/2-iecut+1,isize/2+iecut+1
!           do i=1,isize
           do i=ixmin,ixmax
!            23/06/06 (GL)
             x1=dble(i-isizehalf)-1.0
!            x1=float(i-isize/2)-1.0
!             x1=float(i-isize/2)

            x=xx+x1*pfacX
            ix = min(max(int(x),np),nx(iband)-2-np)
            dx=max(x-ix,0.D0)
            ddx=max(1.D0-dx,0.D0)
 !           write(*,*) "b)",y,np
 !           write(*,*) ix ,iy

            bima0(i,j,ipoint,ns) = (ddx*ddy*sima(ix ,iy,ipoint)   + &
                                    ddx*dy*sima(ix,iy+1,ipoint)   + &
                                    dx* ddy*sima(ix+1 ,iy,ipoint) + &
                                    dx* dy*sima(ix+1,iy+1,ipoint))*fluxfac

          enddo                    !!! end "i=ixmin,ixmax"
         enddo                     !!! end "j=iymin,iymax"
!!!<------
         enddo                     !!! end "iband=1,num_image(ninst)"
         enddo                     !!! end "ninst=1,num_instr"

      endif

      x0old(ns)=x0
      y0old(ns)=y0
    endif

!!!-----------------------------------------------------------------------------


! calculate extent model
! extent model is centered in isize*isize subimage
! needs only be calculated, if parameter 'sig' changes 

  if (sig > sigmin .and. sig /= sigold(ns)) then
!     write(*,*) "ns,sig,sigold:",ns,sig,sigold(ns)
     mima0(:,:,ns)=0.0
!    write(*,*) " calculate ext model"
    if (extmod == "GAUSSIAN") then
     sig4sq=max(sig*sig*16.D0,4.D0)
      do j=1,isizehalf+1
!          23/06/06 (GL)
         y=dble(j-isizehalf-1)
!        y=float(j-isize/2-1)
        do i=1,j
!          23/06/06 (GL)
          x=dble(i-isizehalf-1)
!          x=float(i-isize/2-1)
          if ((x*x+y*y) .lt. sig4sq) then
          if (i/= j) then     
            modval = fgen(0.D0,x-0.5,x+0.5,sig)* &
                     fgen(0.D0,y-0.5,y+0.5,sig)
          else
            modval = fgen(0.D0,x-0.5,x+0.5,sig)
            modval=modval*modval
          endif
            mima0(i,j,ns)=modval
            if (j /= 1) mima0(i,isize-j+2,ns)=modval
            if (j /= 1 .and. i /= 1)  mima0(isize-i+2,isize-j+2,ns)=modval 
            if (i /= 1) then 
              mima0(i,isize-j+2,ns)=modval
              mima0(isize-i+2,j,ns)=modval
            endif
            if (i /= j .and. i /= 1 .and. j /= 1) then
              mima0(j,i,ns)=modval
              if (i /= 1) then
                mima0(j,isize-i+2,ns)=modval
                mima0(isize-j+2,isize-i+2,ns)=modval
              endif
              if (j /= 1) mima0(isize-j+2,i,ns)=modval
            endif
            endif
        end do
      end do    

    else
!!!!    ! beta model
      sig4sq=max(sig*sig*100.D0,5.D0)
      do j=1,isizehalf+1
!        23/06/06 (GL)
        y=dble(j-isizehalf-1)
!        y=float(j-isize/2-1)
        do i=1,j
!        23/06/06 (GL)
          x=dble(i-isizehalf-1)
!          x=float(i-isize/2-1)
          if ((x*x+y*y) .lt. sig4sq) then
            modval=fking(x,y,sig)
            mima0(i,j,ns)=modval
            if (j /= 1) mima0(i,isize-j+2,ns)=modval
            if (j /= 1 .and. i /= 1)  mima0(isize-i+2,isize-j+2,ns)=modval 
            if (i /= 1) then 
              mima0(i,isize-j+2,ns)=modval
              mima0(isize-i+2,j,ns)=modval
            endif
            if (i /= j .and. i /= 1 .and. j /= 1) then
              mima0(j,i,ns)=modval
              if (i /= 1) then
                mima0(j,isize-i+2,ns)=modval
                mima0(isize-j+2,isize-i+2,ns)=modval
              endif
              if (j /= 1) mima0(isize-j+2,i,ns)=modval
           endif
          endif  
        end do
      end do 

!      sig4sq=max(sig*sig*16.,4.)
!       do j=1,isize
!          y=float(j-isize/2)-0.5
!!                          -1.0
!         do i=1,isize
!            x=float(i-isize/2)-0.5
!!                            -1.0
!            if ((x*x+y*y) .lt. sig4sq) then
!              mima0(i,j,ns)=fking(x,y,sig)
!            endif
!         enddo
!       enddo
     norm=sum(mima0(:,:,ns))
!     write(*,*) "sig,norm:",sig,norm  
     if (norm .gt. 1.0) mima0(:,:,ns)=mima0(:,:,ns)/norm
    endif

  endif  
  


!  norm=sum(mima0(:,:,ns))
!  write(*,*) "norm:",sig,norm
!  if (norm .gt. 1.0) mima0(:,:,ns)=mima0(:,:,ns)/norm


! convolve PSF and model using FFTW:


!  write(*,*) "sig:",sig 

  if (sig .gt. sigmin) then
!    write(*,*) "make ffts"
  ! split PSF image in four quadrants (using xsort and ysort) 

    do j=1,isize 
      do i=1,isize
        bima(xsort(i,j),ysort(i,j))=bima0(i,j,iim,ns)
      enddo
    enddo

    call sfftw_execute(pplan)
    mima=mima0(:,:,ns)
    call sfftw_execute(mplan)
    prodima = ffpsf*ffmodel*fftscale
    call sfftw_execute(fplan)
  else  

!!! IT 2012/03 copy only necessary range
!!!    fima = bima0(:,:,iim,ns)
     fima(:,:) = 0
         ixmin=max(1,     isizehalf-iecut+1)
         ixmax=min(isize, isizehalf+iecut)
!         iymin=max(0,     isizehalf-iecut)
!         iymax=min(isize, isizehalf+iecut)
    fima(ixmin:ixmax,ixmin:ixmax) = bima0(ixmin:ixmax,ixmin:ixmax,iim,ns)
 
  endif
!  call cpu_time(t2)
!  call system_clock( clock1,clock2,clock3)
!    write(*,*) "time fftw2:",clock1,clock2,clock3
!  write(*,*) "PSF sum:", sum(fima)

  newcal = .false.
  first  = .false.

!if (sig .gt. sigmin) then
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!1
!      write(*,*) "iim",iim
!      open(unit=1,file='bima0.dat',status='unknown')
!      write(1,*) bima0(:,:,iim,ns)
!      close(1)
!     stop

!      open(unit=1,file='mima01.dat',status='unknown')
!      write(1,*) mima0(:,:,1)
!     close(1)

!      open(unit=1,file='mima02.dat',status='unknown')
!      write(1,*) mima0(:,:,2)
!      close(1)
!      write(*,*) "sig:",sig
!      open(unit=1,file='fima.dat',status='unknown')
!      write(1,*) fima
!     close(1)
!      stop
!endif

  sigold(ns)=sig


  end subroutine fpsf

!---------------------------------------------------------------------------------------

  real(kind=double) function fgen(x0,x1,x2,sig)

    real(kind=double) :: x0,x1,x2,sig,fsig
                                                                       
    fgen=0.
    fsig=sig                                                           
    if (fsig.le.0.00001) fsig=0.00001
    fsig = 1.0/(1.414214*fsig)
    fgen = (ferrf((x2-x0)*fsig)-ferrf((x1-x0)*fsig))*0.5

  end function fgen     

!---------------------------------------------------------------------------------------

  real function ferrf (xx)
 
    real(kind=double) xx
    real t,x

    x=abs(xx)                                                        
    ferrf=1.0
                                                         
    if (x<=5.0) then                                            
       t=1.0/(1.0+0.47047*x)                                            
       ferrf=1.0-(0.3480242-(0.0958798-0.7478556*t)*t)*t*exp(-(x*x))
     end if   
                                                                    
     if (xx<0.0) ferrf=-ferrf                                        

   end function ferrf  

!---------------------------------------------------------------------------------------

     real function fking(x,y,rcore)

!  returns value of the King profile (1+r**2/rcore**2)**(-1.5), averaged over the pixel with center (x,y)
!  normalized to 1  
!  the averageing is approximated by the radial integral between r1 and r2, r1=r_pix-0.5, r2=r_pix+0.5

     real(kind=double) x,y,rcore
     real(kind=double) r1,r2,r1sq,r2sq,rc2,r
     real norm


     rc2=max(1.D-2,rcore*rcore)
  
     r=dsqrt(x*x+y*y)
     r1=max(0.D0,r-0.5D0)
     r2=r+0.5    
     r1sq=r1*r1
     r2sq=r2*r2 

!-----------------------------------------------------
!     kingint=twopi*(rc2/(1+r1sq/rc2)-rc2/(1+r2sq/rc2))
!     area=pi*(r2sq-r1sq)
!     norm=twopi*rc2
!     fking=kingint/area/norm 
!----------------------------------------------------
!     Correct integration of king model (26/06/05 GL):
      norm=twopi*rc2
      fking=(r2/dsqrt(1.+r2sq/rc2)-(r1/dsqrt(1.+r1sq/rc2)))/(r2-r1)/norm
      
 
   end function fking

!---------------------------------------------------------------------------------------


end module fftpsf
